SystemOrganization addCategory: #FixedDecimal!

!Fraction class methodsFor: '*FixedDecimal' stamp: 'cbc 6/13/2005 15:22'!
readFromString: aString
	"Create an object based on the contents of aString."
	| string |

	string := (aString includes: $/) ifTrue: [aString] ifFalse: [aString, '/1'].
	^self readFrom: (ReadStream on: string)! !

!Integer methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (self asFixedDecimal: 0)! !

Number subclass: #FixedDecimal
	instanceVariableNames: 'number scale part1 part2 negative'
	classVariableNames: 'PowersOf10'
	poolDictionaries: ''
	category: 'FixedDecimal'!

!FixedDecimal commentStamp: 'cbc 8/30/2007 20:16' prior: 0!
A FixedDecimal is similar to a ScaledDecimal, but different in certain select ways.  It's primary purpose was to be able to represent precise decimals for such things as representing money - where ScaledDecimals leave something to be desired.  For instance, with ScaledDecimals, you get:
(33.333s withScale:2) + (33.333s withScale:2)   print it yields 66.67s 
but with FixedDecimals, you would get:
(33.333 asFixedDecimal: 2) + (33.333 asFixedDecimal: 2)	print it yields 66.66.
So, FixedDecimals round the numbers to the exact scale you specify - converting a float to a FixedDecimal and back will not necessarily return the starting number, unlike ScaledDecimals.

Instance Variables
	negative:		Indicates wheither the FixedDecimal is negative.  Used for printing internally.
	number:		A raw integer reprsenting the number (not scaled!!).
	part1:		The integer part of the number.
	part2:		The non-integer part of the number (can't remember what that is called right now - too late at night).
	scale:		The scale of the FixedDecimal
!

!FixedDecimal class methodsFor: 'class initialization' stamp: 'cbc 8/31/2010 15:48'!
initialize
	PowersOf10 := Array new: 9.
	PowersOf10 at: 1 put: 10.
	PowersOf10 at: 2 put: 100.
	PowersOf10 at: 3 put: 1000.
	PowersOf10 at: 4 put: 10000.
	PowersOf10 at: 5 put: 100000.
	PowersOf10 at: 6 put: 1000000.
	PowersOf10 at: 7 put: 10000000.
	PowersOf10 at: 8 put: 100000000.
	PowersOf10 at: 9 put: 1000000000.

	super initialize! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/31/2010 16:00'!
newFromFixedDecimal: aDecimal scale: scaleIn 
	"Answer a new instance of me."
	| |

	aDecimal scale = scaleIn ifTrue: [^aDecimal].
	^aDecimal scale < scaleIn
		ifTrue: [
			self basicNew setPart1: aDecimal part1 part2: (aDecimal part2 * (self tenRaisedTo: (scaleIn - aDecimal scale))) scale: scaleIn negative: aDecimal negative
		]  ifFalse: [
			self basicNew setPart1: aDecimal part1 part2: (aDecimal part2 / (self tenRaisedTo: (aDecimal scale - scaleIn))) rounded scale: scaleIn negative: aDecimal negative
		].
! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/31/2010 15:35'!
newFromNumber: aNumber scale: scaleIn 
	"Answer a new instance of me."
	| temp part1 part2 |

	(aNumber isKindOf: FixedDecimal) ifTrue: [^self newFromFixedDecimal: aNumber scale: scaleIn].
	temp := aNumber asFraction.
	part1 := temp truncated.
	part2 := (temp fractionPart * (10 raisedToInteger: scaleIn)) rounded.
	part2 asString size > scaleIn ifTrue: ["Rounding can get us here..."
		temp := (temp * (self tenRaisedTo: scaleIn)) rounded / (10 raisedToInteger: scaleIn).
		part1 := temp truncated.
		part2 := (temp fractionPart * (self tenRaisedTo: scaleIn)) rounded.
		].
	^self basicNew setPart1: part1 part2: part2 scale: scaleIn negative: (temp < 0)! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/13/2010 18:57'!
newFromString: aNumberString scale: scaleIn
	"Answer a new instance of me.
	String is assumed to be of the format: '293492.230' "
	| temp part1 part2 |

	temp := aNumberString readStream.
	part1 := (temp upTo: $.) asSignedInteger.
	temp atEnd ifTrue: [	^self basicNew setPart1: part1 part2: 0 scale: scaleIn negative: (part1 < 0) number: part1].
	part2 := temp next: scaleIn.
	part2 := part2, ((String new: (scaleIn - part2 size)) atAllPut: $0).
	part2 := part2 asInteger.
	^self basicNew setPart1: part1 part2: part2 scale: scaleIn negative: (aNumberString includes: $-)! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'kph 2/2/2008 00:43'!
readFrom: aString
	"compatability with ScaledDecimal"
	
	^ self newFromString: aString scale: (aString size - (aString indexOf: $.))
	
"
FixedDecimal readFrom: '123.456' 
"! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/31/2010 15:35'!
tenRaisedTo: int
	^PowersOf10 at: int ifAbsent: [10 raisedTo: int]! !

!FixedDecimal class methodsFor: 'testing' stamp: 'cbc 8/15/2006 11:38'!
testBoundaryIssues
	self assert: [(FixedDecimal newFromNumber: (407584 * 100 / 5112120) scale: 1) printString = '8.0']! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/22/2006 16:48'!
zero
	^self newFromNumber: 0 scale: 0! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:41'!
* operand 

	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (operand asFraction * self asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (operand * self asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/31/2010 15:34'!
+ operand 
	| nPart1 nPart2 s |
	(negative not and: [operand class = self class and: [operand scale = scale and: [negative = operand negative]]])
		ifTrue: [
			nPart1 := part1 + operand part1.
			nPart2 := part2 + operand part2.
			nPart2 >= (s := self tenRaisedTo: scale) ifTrue: [
				nPart1 := nPart1 + 1.
				nPart2 := nPart2 - s.
				].
			^self class new setPart1: nPart1 part2: nPart2 scale: scale negative: negative number: (number + operand number)
			].
	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (operand asFraction + self asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (operand + self asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/31/2010 15:34'!
- operand 
	| nNumber nPart1 nPart2 nNegative |

	(operand class = self class and: [operand scale = scale]) ifTrue: [
		negative ~= operand negative ifTrue: [
			part1 + operand part1 < (self tenRaisedTo: (scale + 1)) ifTrue: [
				^self class new setPart1: (part1 + operand part1) part2: (part2 + operand part2) scale: scale negative: negative number: (number - operand number)
				] ifFalse: [^self + operand negated].
			].
		nNumber := number - operand number.
		((negative not and: [nNumber < 0]) or: [negative and: [nNumber > 0]]) ifTrue: [nNegative := negative not] ifFalse: [nNegative := negative].
		nPart2 :=  part2 - operand part2.
		nPart1 := part1 - operand part1 - (nPart2 < 0 ifTrue: [1] ifFalse: [0]).
		^self class new setPart1: nPart1 part2: nPart2 scale: scale negative: nNegative number: nNumber
		].
	^((operand isKindOf: FixedDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (self asFraction - operand asFraction) scale: operand scale
	] ifFalse: [
		^self + operand negated
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:42'!
/ operand 

	operand = 0 ifTrue: [^ (ZeroDivide dividend: self) signal].
	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (self asFraction / operand asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (self asFraction / operand asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:40'!
// operand
	"Anser the integer quotient after dividing the receiver by operand
	with truncation towards negative infinity."
	^scale = 0 ifTrue: [part1 // operand] ifFalse: [self asFraction // operand]! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 8/31/2010 15:40'!
< aNumber
	"Implementation of Number 'comparing' method."
	aNumber isNumber ifFalse: [^ false].
	(aNumber isFixedDecimal) ifTrue: [^self lessThanDecimal: aNumber].
	^aNumber adaptToFixedDecimal: self andSend: #<! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 8/31/2010 15:40'!
= aNumber
	"Implementation of Number 'comparing' method."
	aNumber isNumber ifFalse: [^ false].
	(aNumber isFixedDecimal) ifTrue: [^self equalsDecimal: aNumber].
	^aNumber adaptToFixedDecimal: self andSend: #=! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:37'!
adapt: aNumber andSend: selector
	^(aNumber asFixedDecimal: scale) perform: selector with: self! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToFloat: aFloat andSend: selector
	^self adapt: aFloat andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToFraction: aFraction andSend: selector
	^self adapt: aFraction andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToInteger: anInteger andSend: selector
	^self adapt: anInteger andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:39'!
adaptToScaledDecimal: aScaledDecimal andSend: selector
	^self adapt: aScaledDecimal andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:40'!
adaptToString: aString andSend: selector
	^self adapt: aString andSend: selector! !

!FixedDecimal methodsFor: 'truncation and round off' stamp: 'kph 4/1/2009 22:26'!
asFixedDecimal: aScale

	^ self class newFromFixedDecimal: self scale: aScale 
	! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/31/2010 15:33'!
asFloat
	^(number / (self tenRaisedTo: scale)) asFloat! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/31/2010 15:33'!
asFraction
	^number / (self tenRaisedTo: scale)! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'kph 2/1/2009 16:44'!
asInteger
	negative ifTrue: [ ^ part1 negated ].
	
	^ part1
! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 3/11/2005 16:09'!
asScale: newScale
	scale = newScale ifTrue: [^self].
	^self class newFromFixedDecimal: self scale: newScale ! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'kph 2/1/2009 16:44'!
asSignedInteger
	negative ifTrue: [ ^ part1 negated ].
	
	^ part1

! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/31/2010 16:03'!
equalsDecimal: decimal
	| scaleDiff |

	decimal scale = scale ifTrue: [^decimal number = number].
	^(scaleDiff := decimal scale - scale) > 0
		ifTrue: [number * (self tenRaisedTo: scaleDiff) = decimal number]
		ifFalse: [decimal number * (self tenRaisedTo: scaleDiff abs) = number]
! !

!FixedDecimal methodsFor: 'truncation and round off' stamp: 'cbc 9/6/2007 18:59'!
fractionPart
	^FixedDecimal new setPart1: 0 part2: part2 scale: scale negative: negative! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/31/2010 15:33'!
greaterThanDecimal: decimal diff: scaleDiff

	^number * (self tenRaisedTo: (decimal scale - scale)) > decimal number
! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 9/6/2007 18:53'!
hash
	^self asFraction hash! !

!FixedDecimal methodsFor: 'testing' stamp: 'cbc 8/31/2010 15:40'!
isFixedDecimal
	^ true! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/31/2010 16:03'!
lessThanDecimal: decimal
	| scaleDiff |

	decimal scale = scale ifTrue: [^number < decimal number].
	^(scaleDiff := decimal scale - scale) > 0
		ifTrue: [^number * (self tenRaisedTo: scaleDiff) < decimal number]
		ifFalse: [^decimal greaterThanDecimal: self diff: scaleDiff]
! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 9/6/2007 19:05'!
negated
	^FixedDecimal new setPart1: part1 part2: part2 scale: scale negative: negative not! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 8/13/2010 00:46'!
negative
	^negative! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:12'!
number
	^number! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:11'!
part1
	^part1! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:11'!
part2
	^part2! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:58'!
printOn: aStream 
	| p2 |

	negative ifTrue: [aStream nextPut: $-].
	scale = 0 ifTrue: [aStream nextPutAll: part1 asString. ^self].
	p2 := part2 asString.
	p2 := String streamContents: [:s| 
		p2 size + 1 to: scale do: [:x| s nextPut: $0].
		s nextPutAll: p2.
		].
	aStream nextPutAll: part1 asString; nextPut: $.; nextPutAll: p2.! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:59'!
printOn: aStream precision: precise
	| p2 p1 |

	p1 := part1 asString.
	p1 size + scale > precise ifTrue: [^nil].
	p1 := String streamContents: [:s|
		negative ifTrue: [s nextPut: $-].
		p1 size + 1 to: (precise - scale) do: [:x| s nextPut: $0].
		precise - scale > 0 ifTrue: [s nextPutAll: p1].
		].
	scale = 0 ifTrue: [aStream nextPutAll: p1. ^self].
	p2 := part2 asString.
	p2 := String streamContents: [:s| 
		p2 size + 1 to: scale do: [:x| s nextPut: $0].
		s nextPutAll: p2.
		].
	aStream nextPutAll: p1; nextPut: $.; nextPutAll: p2.! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:58'!
printOn: aStream precision: precise reserveSign: boolean
	"ReserveSign, if true, means that if we are not negative, then we should preceed the number with a 0.  So that the numbers will ALWAYS be the same size when printed."
	| p2 p1 |

	p1 := part1 asString.
	p1 size + scale > precise ifTrue: [Error signal: 'Decimal is too big to print!!'].
	negative ifTrue: [aStream nextPut: $-] ifFalse: [boolean ifTrue: [aStream nextPut: $+]].
	p1 size + 1 to: (precise - scale) do: [:x| aStream nextPut: $0].
	precise - scale > 0 ifTrue: [aStream nextPutAll: p1].
	scale = 0 ifTrue: [
		boolean ifTrue: [aStream nextPut: $.].
		^self
		].
	aStream nextPut: $..
	p2 := part2 asString.
	p2 size + 1 to: scale do: [:x| aStream nextPut: $0].
	aStream nextPutAll: p2.
! !

!FixedDecimal methodsFor: 'printing' stamp: 'cbc 10/14/2005 14:40'!
printString
	^String streamContents: [:s| self printOn: s]! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 9/6/2007 18:49'!
reciprocal
	^FixedDecimal newFromNumber: (self asFraction reciprocal) scale: scale! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:12'!
scale
	^scale! !

!FixedDecimal methodsFor: 'printing' stamp: 'ChrisCunningham 2/2/2011 16:36'!
secondsAsHumanDuration
	"Shorten up the number into a rational timeframe that humans like.  1000's of stuff it too much."
	| x |
	
	self > 60 ifFalse: [^self asString, 's'].
	x := self / 60.
	x > 60 ifFalse: [^x asString, 'm'].
	x := x / 60.
	x > 24 ifFalse: [^x asString, 'h'].
	x := x / 24.
	^x asString, 'd'! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/31/2010 15:32'!
setPart1: part1In part2: part2In scale: scaleIn negative: boolean

	part1 := part1In abs.
	part2 := part2In abs.
	scale := scaleIn.
	negative := boolean.
	number := (part1 * (self tenRaisedTo: scale)) + part2.
	negative ifTrue: [number := 0 - number].! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/13/2010 18:37'!
setPart1: part1In part2: part2In scale: scaleIn negative: boolean number: realNumber

	part1 := part1In abs.
	part2 := part2In abs.
	scale := scaleIn.
	negative := boolean.
	number := realNumber! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/31/2010 15:32'!
tenRaisedTo: int
	^PowersOf10 at: int ifAbsent: [10 raisedTo: int]! !

!FixedDecimal methodsFor: 'truncation and round off' stamp: 'cbc 6/13/2005 15:38'!
truncated
	^part1! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 7/5/2011 12:47'!
usedMagnitude
	^part1 asString size! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 7/5/2011 12:54'!
usedScale
	| t x |
	scale < 1 ifTrue: [^0].
	t := (String new: scale) atAllPut: $0.
	x := part2 asString.
	t := (t first: (scale - x size)), x.
	^x reverse asInteger asString size! !

!Number methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (self asFixedDecimal: fixedDecimal scale)! !

!Number methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 19:58'!
asFixedDecimal: scale
	^FixedDecimal newFromNumber: self scale: scale! !

!Number methodsFor: '*FixedDecimal' stamp: 'cbc 8/31/2010 15:39'!
isFixedDecimal
	^ false! !

!ScaledDecimal methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (FixedDecimal newFromNumber: fraction scale: scale)! !

!String methodsFor: '*FixedDecimal' stamp: 'cbc 9/6/2007 19:40'!
asFixedDecimal: scale
	^FixedDecimal newFromString: self scale: scale! !
FixedDecimal initialize!
